from heapq import heappop, heappush
from itertools import count
from typing import Any

import geopandas as gpd
import networkx as nx
import numpy as np
from scipy.spatial import KDTree
from shapely import Geometry


def rivers_dijkstra(graph, source: int, weight: str, cutoff: float):
    push = heappush
    pop = heappop
    weight_remain = {}
    c = count()
    fringe = []
    push(fringe, (next(c), source, cutoff))
    while fringe:
        (_, u, remain_last) = pop(fringe)
        if u in weight_remain:
            weight_remain[u] = weight_remain[u] + remain_last
        else:
            weight_remain[u] = remain_last
        items = graph[u].items()
        match len(items):
            case 1:
                for v, e in items:
                    cost = e.get(weight, 0)
                    if cost is None:
                        continue
                    remain = remain_last - cost
                    if remain < 0:
                        continue
                    push(fringe, (next(c), v, remain))
            case 0:
                _
            case _:
                distrib_weight = {}
                for v, e in items:
                    distrib_weight[v] = _dijkstra_avg_weight(graph, v, weight, remain_last)
                if sum(distrib_weight.values()) == 0:
                    continue
                scale_factor = remain_last / sum(distrib_weight.values())
                distrib_weight = {k: v * scale_factor for k, v in distrib_weight.items()}
                for v, e in items:
                    cost = e.get(weight, 0)
                    if cost is None:
                        continue
                    remain = distrib_weight[v] - cost
                    if remain < 0:
                        continue
                    push(fringe, (next(c), v, remain))
    return weight_remain


def _dijkstra_avg_weight(graph, source: int, weight: str, cutoff: float):
    g_pred = graph._pred
    push = heappush
    pop = heappop
    dist = {}  # dictionary of final distances
    seen = {}
    c = count()
    fringe = []
    seen[source] = 0
    push(fringe, (0, next(c), source))
    while fringe:
        (d, _, u) = pop(fringe)
        if u in dist:
            continue  # already searched this node.
        dist[u] = d
        items = graph[u].items()
        if len(items) > 1 or len(g_pred[u].items()) > 1:
            continue
        for v, e in items:
            cost = e.get(weight, 0)
            if cost is None:
                continue
            uv_dist = d + cost
            if uv_dist > cutoff:
                continue
            if v not in seen or uv_dist < seen[u]:
                seen[v] = uv_dist
                push(fringe, (uv_dist, next(c), v))

    return np.average(list(seen.values()))


def get_closest_nodes(gdf_from: gpd.GeoDataFrame, to_nx_graph: nx.Graph) -> list[tuple[Any, Any]]:
    if gdf_from.crs.to_epsg() != to_nx_graph.graph["crs"]:
        gdf_from.to_crs(to_nx_graph.graph["crs"], inplace=True)

    mapping = dict((u, id) for (id, u) in zip(to_nx_graph.nodes(), range(to_nx_graph.number_of_nodes())))
    points = gdf_from.representative_point()
    coordinates = [(data["x"], data["y"]) for node, data in to_nx_graph.nodes(data=True)]
    tree = KDTree(coordinates)
    target_coord = [(p.x, p.y) for p in points]
    distances, indices = tree.query(target_coord)
    return [(mapping.get(index), dist) for dist, index in zip(distances, indices)]


def graph_to_gdf(
    graph: nx.MultiDiGraph,
) -> gpd.GeoDataFrame | None:
    """
    Converts nx graph to gpd.GeoDataFrame as edges.

    Parameters
    ----------
    graph : nx.MultiDiGraph
        The graph to convert.

    Returns
    -------
    gpd.GeoDataFrame
        Graph representation in GeoDataFrame format
    """

    crs = graph.graph["crs"]
    e_ind_source, e_ind_target, e_data = zip(*graph.edges(data=True))
    index_matrix = np.array([e_ind_source, e_ind_target]).transpose()
    final_index = [tuple(i) for i in index_matrix]
    lines = ((d["geometry"]) if isinstance(d["geometry"], Geometry) else None for d in e_data)
    gdf_edges = gpd.GeoDataFrame(e_data, index=final_index, crs=crs, geometry=list(lines))
    return gdf_edges



import math
import os

import dask.dataframe as dd
import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
from dask.diagnostics import ProgressBar
from loguru import logger
from shapely import LineString, MultiLineString, Point, convex_hull, unary_union
from tqdm.auto import tqdm


def construct_water_graph(
    rivers: gpd.GeoDataFrame,
    water: gpd.GeoDataFrame,
    segmentize_len=300,
    max_river_width=50000,
    basic_river_width=5,
    basic_river_speed=50,
) -> nx.DiGraph:
    """
    Constructs a directed graph representing a river network, with nodes and edges defined by river geometry and water bodies.

    This function creates nodes and edges from river line geometries, assigns attributes to each edge such as width
    and flow speed, and returns the river network as a directed graph. The function also takes into account the
    proximity of water bodies to rivers to determine edge widths and geometries.

    Args:
        rivers (gpd.GeoDataFrame): GeoDataFrame containing the river geometries as LineString objects.
        water (gpd.GeoDataFrame): GeoDataFrame containing the water body geometries, used to adjust river widths
            where rivers intersect water bodies.
        segmentize_len (int, optional): Length for segmenting the river lines, used to split long lines into smaller
            segments. Default is 300.
        max_river_width (int, optional): Maximum river width. Used in calculating edge width based on proximity
            to water bodies. Default is 50000.
        basic_river_width (int, optional): Base width for rivers when no other data is available. Default is 5.
        basic_river_speed (int, optional): Base speed for calculating edge weight. Default is 50.

    Returns:
        nx.DiGraph: A directed graph representing the river network, with nodes and edges. Edge attributes include
        `width`, `length`, `weight`, and `geometry`. Node attributes include `x` and `y` coordinates.

    """
    local_crs = rivers.estimate_utm_crs()
    rivers.to_crs(local_crs, inplace=True)

    if water.crs != rivers.crs:
        water.to_crs(rivers.crs, inplace=True)

    rivers["geometry"] = rivers.geometry
    water["geometry"] = water.geometry
    rivers.set_geometry("geometry", inplace=True)
    water.set_geometry("geometry", inplace=True)
    node_id = 0
    nodes = []
    edges = []

    def distance(x1, y1, x2, y2):
        return math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)

    def add_node(node_id_to_add, x, y):
        nodes.append({"node_id": node_id_to_add, "point": (x, y)})

    def add_edge(u, v, length, ux, uy, vx, vy):
        edges.append({"u": u, "v": v, "length": length, "ux": ux, "uy": uy, "vx": vx, "vy": vy})

    def process_edge(p1, p2):
        nonlocal node_id
        add_node(node_id, p1[0], p1[1])
        node_id += 1
        add_edge(node_id - 1, node_id, distance(p1[0], p1[1], p2[0], p2[1]), p1[0], p1[1], p2[0], p2[1])
        add_node(node_id, p2[0], p2[1])
        node_id += 1

    def explode_linestrings_to_edgenodes(loc):
        coords = loc.geometry.segmentize(segmentize_len).coords
        for i, _ in enumerate(coords[:-1]):
            process_edge(coords[i], coords[i + 1])

    rivers.apply(explode_linestrings_to_edgenodes, axis=1)
    logger.debug("Done exploding rivers!")
    nodes, edges = pd.DataFrame(nodes), pd.DataFrame(edges)
    grouped_nodes = nodes.groupby("point")["node_id"].agg(list).reset_index()
    logger.debug("Done grouping nodes by coordinates!")
    mapping = {node_id: idx for idx, node_ids in enumerate(grouped_nodes["node_id"]) for node_id in node_ids}

    edges["u"] = edges["u"].map(mapping)
    edges["v"] = edges["v"].map(mapping)
    logger.debug("Done remapping edges!")
    grouped_nodes_gdf = gpd.GeoDataFrame(geometry=[Point(x[0], x[1]) for x in grouped_nodes["point"]], crs=local_crs)
    grouped_nodes = grouped_nodes.drop(columns="node_id").reset_index(names="node_id")
    rivers_poly = gpd.sjoin(water, grouped_nodes_gdf, how="inner", predicate="intersects")
    points_within_water = rivers_poly["index_right"].to_list()

    rivers_poly_v = rivers_poly[["index_right", "geometry"]].drop_duplicates(subset="index_right")

    edgenodes = (
        pd.merge(edges, rivers_poly_v, how="left", left_on="v", right_on="index_right")
        .drop(columns=["index_right"])
        .rename({"geometry": "geometry_water_v"}, axis=1)
    )
    not_visited_nodes = list(
        set(edgenodes[edgenodes["geometry_water_v"].notna()]["v"].tolist()) ^ set(points_within_water)
    )
    rivers_poly_u = rivers_poly_v[rivers_poly_v["index_right"].isin(not_visited_nodes)]
    edgenodes = (
        pd.merge(edgenodes, rivers_poly_u, how="left", left_on="u", right_on="index_right")
        .drop(columns=["index_right"])
        .rename({"geometry": "geometry_water_u"}, axis=1)
    )
    logger.debug("Calculating river's widths...")
    dask_edgenodes = dd.from_pandas(edgenodes, npartitions=os.cpu_count() * 2 + 1)
    dask_edgenodes["geometries"] = dask_edgenodes.apply(
        lambda x: _calc_edge_width(x, max_dist=max_river_width, basic_dist=basic_river_width), axis=1, meta=tuple
    )
    with ProgressBar():
        edgenodes = dask_edgenodes.compute()

    edgenodes[["geometry_v", "geometry_u"]] = pd.DataFrame(edgenodes["geometries"].tolist(), index=edgenodes.index)
    grouped_nodes = grouped_nodes.merge(
        edgenodes[["v", "geometry_v"]], how="left", left_on="node_id", right_on="v"
    ).drop(columns="v")
    grouped_nodes = grouped_nodes.merge(
        edgenodes[["u", "geometry_u"]], how="left", left_on="node_id", right_on="u"
    ).drop(columns="u")
    grouped_nodes["geometry"] = grouped_nodes["geometry_v"].combine_first(grouped_nodes["geometry_u"])
    grouped_nodes.drop(columns=["geometry_v", "geometry_u"], inplace=True)
    grouped_nodes["width"] = grouped_nodes["geometry"].apply(
        lambda x: x.length if isinstance(x, LineString) else np.inf
    )
    grouped_nodes = grouped_nodes.loc[grouped_nodes.groupby("node_id")["width"].idxmin()]
    edges = (
        edges.merge(grouped_nodes[["node_id", "width", "geometry"]], left_on="u", right_on="node_id")
        .drop(columns="node_id")
        .rename(columns={"width": "width_u", "geometry": "geometry_u"})
    )
    edges = (
        edges.merge(grouped_nodes[["node_id", "width", "geometry"]], left_on="v", right_on="node_id")
        .drop(columns="node_id")
        .rename(columns={"width": "width_v", "geometry": "geometry_v"})
    )
    edges.replace(np.inf, basic_river_width, inplace=True)
    grouped_nodes[["x", "y"]] = pd.DataFrame(grouped_nodes["point"].tolist(), index=grouped_nodes.index)
    grouped_nodes.reset_index(drop=True, inplace=True)

    graph = nx.DiGraph()
    for _, edge in tqdm(edges.iterrows(), total=len(edges), desc="Processing edges"):
        u = int(edge["u"])
        v = int(edge["v"])
        width = (edge["width_u"] + edge["width_v"]) / 2
        geometry_u = edge["geometry_u"]
        geometry_v = edge["geometry_v"]
        if pd.isna(geometry_u) and pd.isna(geometry_v):
            geometry = None
        else:
            if pd.isna(geometry_u):
                geometry_u = Point(edge["ux"], edge["uy"])
            if pd.isna(geometry_v):
                geometry_v = Point(edge["vx"], edge["vy"])
            geometry = convex_hull(unary_union([geometry_u, geometry_v]))
        graph.add_edge(
            u,
            v,
            width=width,
            length=edge["length"],
            weight=width * edge["length"] / basic_river_speed,
            geometry=geometry,
        )

    graph.add_nodes_from(set(grouped_nodes.index) - set(graph.nodes))
    for col in grouped_nodes[["x", "y"]].columns:
        nx.set_node_attributes(graph, name=col, values=grouped_nodes[col].dropna().astype(np.float32))
    graph.graph["crs"] = local_crs.to_epsg()
    return graph


def _calc_edge_width(loc, max_dist, basic_dist):
    def perpendicular_to_line_end(p1, p2, length: float):
        angle = math.atan2(p2[1] - p1[1], p2[0] - p1[0])
        perp_angle = angle + math.pi / 2

        dx = (length / 2) * math.cos(perp_angle)
        dy = (length / 2) * math.sin(perp_angle)

        perp_start = (p2[0] - dx, p2[1] - dy)
        perp_end = (p2[0] + dx, p2[1] + dy)
        return LineString([perp_start, p2, perp_end])

    def trim_line_by_polygon(perpendicular, river_polygon, mid_point):
        def get_line_containing_point(multilinestring, point):
            for line in multilinestring.geoms:
                if line.contains(point):
                    return line
            return None

        intersection = perpendicular.intersection(river_polygon)
        match intersection:
            case _ if intersection.is_empty:
                return None
            case LineString():
                return intersection
            case MultiLineString():
                return get_line_containing_point(intersection, mid_point)
            case _:
                return None

    if loc.geometry_water_u is None and loc.geometry_water_v is None:
        return perpendicular_to_line_end((loc.ux, loc.uy), (loc.vx, loc.vy), basic_dist), perpendicular_to_line_end(
            (loc.vx, loc.vy), (loc.ux, loc.uy), basic_dist
        )
    cutted_perp_uv = None
    cutted_perp_vu = None

    if loc.geometry_water_v is not None:
        perp_uv = perpendicular_to_line_end((loc.ux, loc.uy), (loc.vx, loc.vy), max_dist)
        cutted_perp_uv = trim_line_by_polygon(perp_uv, loc.geometry_water_v, Point((loc.vx, loc.vy)))

    if loc.geometry_water_u is not None:
        perp_vu = perpendicular_to_line_end((loc.vx, loc.vy), (loc.ux, loc.uy), max_dist)
        cutted_perp_vu = trim_line_by_polygon(perp_vu, loc.geometry_water_u, Point((loc.ux, loc.uy)))

    if cutted_perp_uv is None:
        cutted_perp_uv = perpendicular_to_line_end((loc.ux, loc.uy), (loc.vx, loc.vy), basic_dist)

    if cutted_perp_vu is None:
        cutted_perp_vu = perpendicular_to_line_end((loc.vx, loc.vy), (loc.ux, loc.uy), basic_dist)
    return cutted_perp_uv, cutted_perp_vu



import geopandas as gpd
import matplotlib
from matplotlib import pyplot as plt

from ecodonut.eco_frame import EcoFrame


def value_to_color(
    eco_frame: EcoFrame, color_column, negative_minmax: tuple = (-10, 0), positive_minmax: tuple = (0, 4)
) -> EcoFrame:
    cmapYlGn = plt.get_cmap("YlGn")
    cmapYlOrRd = plt.get_cmap("YlOrRd")

    norm_pos = plt.Normalize(*positive_minmax)
    norm_neg = plt.Normalize(*negative_minmax)

    eco_frame[color_column] = eco_frame["layer_impact"].apply(
        lambda x: cmapYlGn(norm_pos(x)) if x >= 0 else cmapYlOrRd(norm_neg(x))
    )
    eco_frame[color_column] = eco_frame[color_column].apply(matplotlib.colors.to_hex)
    return eco_frame


def get_map(ecoframe: EcoFrame, tiles="CartoDB positron"):

    ecoframe = value_to_color(ecoframe, "color")
    m = gpd.GeoDataFrame(ecoframe, geometry="geometry", crs=ecoframe.local_crs).explore(
        color=ecoframe["color"],
        tiles=tiles,
    )
    return m



from .utils import (
    calc_layer_count,
    combine_geometry,
    create_buffers,
    merge_objs_by_buffer,
    min_max_normalization,
    project_points_into_polygons,
    polygons_to_linestring,
)
from .graph_utils import rivers_dijkstra, get_closest_nodes, graph_to_gdf



import math
from typing import Callable, Dict, Tuple

import geopandas as gpd
import numpy as np
import pandas as pd
from loguru import logger
from shapely.ops import unary_union

from ecodonut.eco_frame.eco_layers import LayerOptions, default_layers_options
from ecodonut.utils import calc_layer_count, combine_geometry, create_buffers, merge_objs_by_buffer


def _positive_fading(layers_count: int, i: int) -> float:
    """Calculate fading effect for positive layers."""
    sigmoid_value = math.exp(-(i - 0.5) * ((0.7 * math.e**2) / layers_count))
    return sigmoid_value


def _negative_fading(layers_count: int, i: int) -> float:
    """Calculate fading effect for negative layers."""
    sigmoid_value = 1 / (1 + math.exp(10 / layers_count * (i - 0.5 - (layers_count / 2))))
    return sigmoid_value


def _calculate_impact(impact_list: list) -> float:
    if len(impact_list) == 1:
        return impact_list[0]
    positive_list = sorted([x for x in impact_list if x > 0])
    negative_list = sorted([abs(x) for x in impact_list if x < 0])
    total_positive = 0
    for imp in positive_list:
        total_positive = np.sqrt(imp**2 + total_positive**2)

    total_negative = 0
    for imp in negative_list:
        total_negative = np.sqrt(imp**2 + total_negative**2)

    return total_positive - total_negative


class EcoFrame(gpd.GeoDataFrame):
    """
    A GeoDataFrame-based class that represents an ecological frame (eco-frame),
    storing geometry and associated ecological data with customizable settings.
    """

    _metadata = ["min_donut_count_radius", "max_donut_count_radius", "negative_types", "positive_types", "local_crs"]

    def __init__(self, min_donut_count_radius, max_donut_count_radius, negative_types, positive_types, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.negative_types = negative_types
        self.positive_types = positive_types
        self.min_donut_count_radius = min_donut_count_radius
        self.max_donut_count_radius = max_donut_count_radius
        self.local_crs = self.crs

    @property
    def _constructor(self):
        return EcoFrame

    def __finalize__(self, other, method=None, **kwargs):
        self = super().__finalize__(other, method=method, **kwargs)
        for attr in self._metadata:
            setattr(self, attr, getattr(other, attr, None))
        return self


class EcoFrameCalculator:
    """
    A class for calculating eco-frame based on geographic layers and customizable parameters.

    Attributes:
        max_donut_count_radius (float): Maximum radius for donut count calculation.
        min_donut_count_radius (float): Minimum radius for donut count calculation.
        local_crs (CRS): Local coordinate reference system for the territory.
        territory (GeoDataFrame): GeoDataFrame of the target territory.
        layer_options (Dict[str, LayerOptions]): Layer options for eco-frame evaluation.
        positive_fading_func (Callable): Function to calculate positive fading effect.
        negative_fading_func (Callable): Function to calculate negative fading effect.
        impact_calculator (Callable): Function to calculate the impact score.
    """

    max_donut_count_radius = None
    min_donut_count_radius = None

    def __init__(
        self,
        territory: gpd.GeoDataFrame,
        settings_from: EcoFrame = None,
        layer_options: Dict[str, LayerOptions] = None,
        positive_fading_func: Callable[[int, int], float] = _positive_fading,
        negative_fading_func: Callable[[int, int], float] = _negative_fading,
        impact_calculator: Callable[[Tuple[float, ...]], float] = _calculate_impact,
    ):
        """
        Initializes the EcoFrameCalculator with specified settings.

        Args:
            territory (GeoDataFrame): Target territory for eco-frame calculations.
            settings_from (EcoFrame, optional): Existing EcoFrame for inheriting settings.
            layer_options (Dict[str, LayerOptions], optional): Layer options for the eco-frame.
            positive_fading_func (Callable, optional): Function to compute positive fading.
            negative_fading_func (Callable, optional): Function to compute negative fading.
            impact_calculator (Callable, optional): Function to compute overall impact score.
        """
        if layer_options is None:
            layer_options = default_layers_options
        self.local_crs = territory.estimate_utm_crs()
        self.territory = territory.to_crs(self.local_crs)
        if settings_from is not None:
            self.max_donut_count_radius = settings_from.max_donut_count_radius
            self.min_donut_count_radius = settings_from.min_donut_count_radius
        self.layer_options = layer_options
        self.positive_fading_func = positive_fading_func
        self.negative_fading_func = negative_fading_func
        self.impact_calculator = impact_calculator

    def evaluate_ecoframe(
        self,
        eco_layers: Dict[str, gpd.GeoDataFrame],
        min_layer_count: int = 2,
        max_layer_count: int = 10,
    ) -> EcoFrame:
        """
        Creates an EcoFrame from specified ecological layers.

        Args:
            eco_layers (Dict[str, GeoDataFrame]): Dictionary of ecological layers by name.
            min_layer_count (int, optional): Minimum count of layers to include in donut calculation.
            max_layer_count (int, optional): Maximum count of layers to include in donut calculation.

        Returns:
            EcoFrame: Generated EcoFrame instance containing ecological impact data.
        """
        layers_to_donut = []
        backgrounds = []
        positive_layers = {}
        negative_layers = {}
        for layer_name, cur_eco_layer in eco_layers.items():
            if cur_eco_layer is None:
                continue
            if layer_name not in self.layer_options:
                raise RuntimeError(f"{layer_name} is not provided in layer options")
            logger.debug(f'Processing layer "{layer_name}"')
            cur_eco_layer = cur_eco_layer.copy()

            layer_options = self.layer_options[layer_name]
            cur_eco_layer.to_crs(self.local_crs, inplace=True)

            # apllying geometry functions
            geom_func = layer_options.geom_func
            if geom_func is not None:
                cur_eco_layer.geometry = cur_eco_layer.geometry.apply(geom_func)
            cur_eco_layer = cur_eco_layer[cur_eco_layer.geom_type.isin(["MultiPolygon", "Polygon"])]

            # merging geometry if indeed
            russian_name = layer_options.russian_name
            union, union_buff = layer_options.geom_union_unionbuff
            if union:
                cur_eco_layer.geometry = cur_eco_layer.geometry.simplify(layer_options.simplify)
                cur_eco_layer = cur_eco_layer.geometry.buffer(union_buff, 8).union_all().buffer(-union_buff, 8)
                cur_eco_layer = gpd.GeoDataFrame(geometry=[cur_eco_layer], crs=self.local_crs)
                cur_eco_layer["name"] = russian_name

            # filling "name" column
            if "name" not in cur_eco_layer.columns:
                cur_eco_layer["name"] = f"{russian_name} без названия"
            cur_eco_layer.fillna({"name": f"{russian_name} без названия"}, inplace=True)

            # calculation INITIAL_IMPACT coefficient for layer
            initial_impact = layer_options.initial_impact

            cur_eco_layer["initial_impact"] = (
                initial_impact
                if isinstance(initial_impact, int)
                else (cur_eco_layer[initial_impact[0]].apply(initial_impact[1]))
            )
            if (cur_eco_layer["initial_impact"] > 0).all():
                positive_layers.update({layer_name: russian_name})
            else:
                negative_layers.update({layer_name: russian_name})

            # calculation Fading coefficient for layer
            fading = layer_options.fading
            cur_eco_layer["fading"] = (
                fading if isinstance(fading, float) else (cur_eco_layer[fading[0]].apply(fading[1]))
            )

            # calculation area_normalization coefficient for layer
            area_normalization = layer_options.area_normalization
            cur_eco_layer["area_normalization"] = (
                1 if area_normalization is None else (area_normalization(cur_eco_layer.area))
            )

            # calculation total_impact_radius in meters for layer
            cur_eco_layer["total_impact_radius"] = np.round(
                cur_eco_layer["initial_impact"] * cur_eco_layer["fading"] * cur_eco_layer["area_normalization"] * 1000,
                1,
            )

            cur_eco_layer["geometry"] = cur_eco_layer.geometry

            # merging objects if merge_radius
            merge_buff = layer_options.merge_radius
            if merge_buff is not None:
                cur_eco_layer = merge_objs_by_buffer(cur_eco_layer, merge_buff)
                cur_eco_layer["initial_impact"] = cur_eco_layer["initial_impact"].apply(
                    lambda x: (
                        min(x)
                        if isinstance(x, tuple) and all(i < 0 for i in x)
                        else max(x) if isinstance(x, tuple) and all(i > 0 for i in x) else x
                    )
                )

                cur_eco_layer["total_impact_radius"] = cur_eco_layer["total_impact_radius"].apply(
                    lambda x: (
                        min(x)
                        if isinstance(x, tuple) and all(i < 0 for i in x)
                        else max(x) if isinstance(x, tuple) and all(i > 0 for i in x) else x
                    )
                )
            cur_eco_layer["type"] = layer_name
            cur_eco_layer.geometry = cur_eco_layer.geometry.simplify(layer_options.simplify)
            if layer_options.make_donuts:
                layers_to_donut.append(cur_eco_layer)
            else:
                cur_eco_layer["layer_impact"] = cur_eco_layer["initial_impact"]
                cur_eco_layer["source"] = True
                backgrounds.append(cur_eco_layer)
        logger.debug("Donutting layers...")
        donuted_layers = gpd.GeoDataFrame(
            pd.concat(layers_to_donut, ignore_index=True), geometry="geometry", crs=self.local_crs
        )

        donuted_layers = donuted_layers.filter(
            items=["name", "type", "initial_impact", "total_impact_radius", "geometry"]
        )

        logger.debug("Calculation layer's count...")
        if self.min_donut_count_radius is None and self.max_donut_count_radius is None:
            donuted_layers["layers_count"] = calc_layer_count(donuted_layers, min_layer_count, max_layer_count)
            impacts = np.abs(donuted_layers["total_impact_radius"])
            min_layer_rad = int(np.min(impacts))
            max_layer_rad = int(np.max(impacts))
        else:
            min_layer_count, min_layer_rad = self.min_donut_count_radius
            max_layer_count, max_layer_rad = self.max_donut_count_radius
            donuted_layers["layers_count"] = calc_layer_count(
                donuted_layers, min_layer_count, max_layer_count, min_layer_rad, max_layer_rad
            )

        min_donut_count_radius = min_layer_count, min_layer_rad
        max_donut_count_radius = max_layer_count, max_layer_rad
        logger.debug("Distributing levels...")
        donuted_layers = self._distribute_levels(donuted_layers)
        donuted_layers = pd.concat([donuted_layers] + backgrounds, ignore_index=True)
        logger.debug("Combining geometry...")
        donuted_layers = combine_geometry(donuted_layers, self.impact_calculator)
        donuted_layers = donuted_layers.clip(self.territory, keep_geom_type=True)
        logger.debug("Grouping geometry...")
        donuted_layers = (
            donuted_layers.groupby(["name", "layer_impact"])
            .agg({"type": "first", "source": "first", "geometry": lambda x: unary_union(x)})
            .reset_index()
        )
        donuted_layers = EcoFrame(
            min_donut_count_radius=min_donut_count_radius,
            max_donut_count_radius=max_donut_count_radius,
            positive_types=positive_layers,
            negative_types=negative_layers,
            data=donuted_layers,
            geometry="geometry",
            crs=self.local_crs,
        )
        return donuted_layers

    def _distribute_levels(self, data: gpd.GeoDataFrame, resolution=4) -> gpd.GeoDataFrame:
        distributed = data.copy().apply(
            create_buffers,
            resolution=resolution,
            positive_func=self.positive_fading_func,
            negative_func=self.negative_fading_func,
            axis=1,
        )
        return gpd.GeoDataFrame(
            pd.concat(distributed.tolist(), ignore_index=True), geometry="geometry", crs=self.local_crs
        )


class TerritoryMark:
    absolute_mark: float = None
    absolute_mark_description: str = None
    relative_mark: float = None
    relative_mark_description: str = None
    clipped_ecoframe: gpd.GeoDataFrame = None

    def __init__(
        self,
        absolute_mark: float,
        absolute_mark_description: str,
        relative_mark: float,
        relative_mark_description: str,
        clipped_ecoframe: gpd.GeoDataFrame,
    ):
        self.absolute_mark = absolute_mark
        self.absolute_mark_description = absolute_mark_description
        self.relative_mark = relative_mark
        self.relative_mark_description = relative_mark_description
        self.clipped_ecoframe = clipped_ecoframe

    def __str__(self):
        area = self.clipped_ecoframe.geometry.area.sum() if self.clipped_ecoframe is not None else "Неизвестно"
        return (
            f"Полная оценка территории:\n"
            f"  Абсолютная оценка: {self.absolute_mark}\n"
            f"  Описание объектов: {self.absolute_mark_description}\n\n"
            f"  Относительная оценка: {self.relative_mark}\n"
            f"  Интерпретация оценки: {self.relative_mark_description}\n\n"
            f"  Площадь: {area}\n"
        )

    def __repr__(self):
        return self.__str__()


def mark_territory(eco_frame: EcoFrame, zone: gpd.GeoDataFrame = None) -> TerritoryMark:
    """
    Generates a territory mark by assessing ecological impact within a specified zone.

    Args:
        eco_frame (EcoFrame): The eco-frame containing ecological layers.
        zone (GeoDataFrame, optional): Specific area to calculate ecological impact for.

    Returns:
        TerritoryMark: Calculated territory mark object containing impact assessment.
    """
    zone = zone.copy()
    eco_frame.set_geometry("geometry", inplace=True)
    eco_frame.to_crs(eco_frame.local_crs, inplace=True)
    zone.to_crs(eco_frame.local_crs, inplace=True)
    if zone is None:
        clip = eco_frame.copy()

        total_area = sum(clip.geometry.area)
    else:
        clip = eco_frame.clip(zone)

        total_area = sum(clip.geometry.area)
    if clip.empty:
        desc = "В границах проектной территории нет данных об объектах оказывающих влияние на экологию"
        obj_msg = "В границы проектной территории не попадает влияние от обьектов, оказывающих влияние на экологию."
        return TerritoryMark(0, obj_msg, 0, desc, clip)
    negative_types = list(eco_frame.negative_types.keys())
    clip["impact_percent"] = clip.geometry.area / total_area
    abs_mark = round(sum(clip["layer_impact"] * clip["impact_percent"]), 2)

    unique_sources = pd.DataFrame(
        clip.apply(lambda x: list(zip(x["name"], x["type"], x["source"])), axis=1).explode().unique().tolist(),
        columns=["name", "type", "source"],
    )
    negative_sources = unique_sources[unique_sources["source"]]
    negative_sources = negative_sources[negative_sources["type"].isin(negative_types)]

    negative_effectors = unique_sources[~unique_sources["source"]]
    negative_effectors = negative_effectors[negative_effectors["type"].isin(negative_types)]
    negative_effectors = negative_effectors[~negative_effectors["name"].isin(negative_sources["name"])]

    if len(negative_sources) > 0:
        obj_message = (
            f"На проектной территории находятся {len(negative_sources)} источник(а/ов) негативного воздействия:"
        )
        for ind, source in negative_sources.reset_index(drop=True).iterrows():
            name = source["name"]
            russian_name = eco_frame.negative_types.get(source["type"])
            if isinstance(name, str):
                if russian_name in name:
                    obj_message += f"\n{ind + 1}. {name}"
                else:
                    obj_message += f'\n{ind + 1}. {russian_name}: "{name}"'
            if isinstance(name, tuple):
                formatted_names = "\n    ".join(f"- {item}" for item in name)
                obj_message += f'\n{ind + 1}. Множество объектов типа "{russian_name}":\n    {formatted_names}'
    else:
        obj_message = "\nИсточников негативного воздействия на проектной территории нет."

    if len(negative_effectors) > 0:
        obj_message += f"\n\nВ проектную территорию попадает влияние от {len(negative_effectors)} источник(а/ов) негативного воздействия:"
        for ind, source in negative_effectors.reset_index(drop=True).iterrows():
            name = source["name"]
            russian_name = eco_frame.negative_types.get(source["type"])
            if isinstance(name, str):
                if russian_name in name:
                    obj_message += f"\n{ind + 1}. {name}"
                else:
                    obj_message += f'\n{ind + 1}. {russian_name}: "{name}"'
            if isinstance(name, tuple):
                formatted_names = "\n    ".join(f"- {item}" for item in name)
                obj_message += f'\n{ind + 1}. Множество объектов типа "{russian_name}":\n    {formatted_names}'

    if (clip["layer_impact"] > 0).all():
        desc = (
            "Проектная территория имеет оценку 5 баллов по экологическому каркасу. Территория находится в зоне "
            "влияния объектов, оказывающих только положительное влияние на окружающую среду."
        )
        return TerritoryMark(abs_mark, obj_message, 5, desc, clip)

    if (clip["layer_impact"] < 0).all():
        desc = (
            "Проектная территория имеет оценку 0 баллов по экологическому каркасу. Территория находится в зоне "
            "влияния объектов, оказывающих только отрицательное влияние на окружающую среду."
        )
        return TerritoryMark(abs_mark, obj_message, 0, desc, clip)

    if abs_mark >= 2:
        desc = (
            "Проектная территория имеет оценку 4 балла по экологическому каркасу. Территория находится "
            "преимущественно в зоне влияния объектов, оказывающих положительное влияние на окружающую среду."
        )

        return TerritoryMark(abs_mark, obj_message, 4, desc, clip)

    if abs_mark > 0:
        desc = (
            "Проектная территория имеет оценку 3 балла по экологическому каркасу. Территория находится в зоне влияния "
            "как положительных, так и отрицательных объектов, однако положительное влияние оказывает большее "
            "воздействие чем отрицательное."
        )

        return TerritoryMark(abs_mark, obj_message, 3, desc, clip)

    if abs_mark == 0:
        desc = (
            "Проектная территория имеет оценку 2.5 балла по экологическому каркасу. Территория находится в зоне "
            "влияния как положительных, так и отрицательных объектов, однако положительные и негативные влияния "
            "компенсируют друг друга."
        )

        return TerritoryMark(abs_mark, obj_message, 2.5, desc, clip)

    if abs_mark >= -4:
        desc = (
            "Проектная территория имеет оценку 2 балла по экологическому каркасу. Территория находится в зоне влияния "
            "как положительных, так и отрицательных объектов, однако отрицательное влияние оказывает большее "
            "воздействие чем положительное."
        )

        return TerritoryMark(abs_mark, obj_message, 2, desc, clip)

    if abs_mark < -4:
        desc = (
            "Проектная территория имеет оценку 1 балл по экологическому каркасу. Территория находится "
            "преимущественно в зоне влияния объектов, оказывающих негативное влияние на окружающую среду."
        )

        return TerritoryMark(abs_mark, obj_message, 1, desc, clip)


def concat_ecoframes(eco_frame1: EcoFrame, eco_frame2: EcoFrame, impact_calculator=_calculate_impact) -> EcoFrame:
    """
    Merges two eco-frames into a single EcoFrame with combined geometries and impacts.

    Args:
        eco_frame1 (EcoFrame): First eco-frame to merge.
        eco_frame2 (EcoFrame): Second eco-frame to merge.
        impact_calculator (Callable): Function to calculate impact during merging.

    Returns:
        EcoFrame: Merged EcoFrame containing combined ecological data and geometries.
    """
    frame1 = eco_frame1.copy()
    frame2 = eco_frame2.copy()

    negative_types = eco_frame1.negative_types.copy()
    negative_types.update(eco_frame2.negative_types)

    positive_types = eco_frame1.positive_types.copy()
    positive_types.update(eco_frame2.positive_types)

    ind_to_change = frame1.sjoin(frame2).index.unique()

    new_frame = pd.concat([frame1.loc[ind_to_change], frame2], ignore_index=True)
    new_frame = combine_geometry(new_frame, impact_calculator, "sum")
    new_frame = pd.concat([frame1.drop(ind_to_change), new_frame], ignore_index=True)
    new_frame = (
        new_frame.groupby(["name", "layer_impact"])
        .agg({"type": "first", "source": "first", "geometry": unary_union})
        .reset_index()
    )
    new_frame = EcoFrame(
        min_donut_count_radius=eco_frame1.min_donut_count_radius,
        max_donut_count_radius=eco_frame1.max_donut_count_radius,
        positive_types=negative_types,
        negative_types=negative_types,
        data=new_frame,
        geometry="geometry",
        crs=frame1.crs,
    )
    return new_frame



import math
from dataclasses import dataclass
from typing import Any, Callable, Dict, Tuple

import numpy as np
from numpy import ndarray
from shapely.geometry.base import BaseGeometry

from ecodonut.utils import min_max_normalization


@dataclass
class LayerOptions:
    initial_impact: Tuple[str, Callable[[Any], int]] | int
    fading: Tuple[str, Callable[[Any], float]] | float
    russian_name: str
    geom_func: Callable[[BaseGeometry], BaseGeometry] | None = None
    area_normalization: Callable[[ndarray], ndarray] | None = None
    make_donuts: bool = True

    # If first is True will union all layer geometries,also leads to unique 'name' attribute loose,
    # second value stands for union buffer, it actually will do geometry.buffer(val).unary_union.buffer(-val)
    geom_union_unionbuff: Tuple[bool, int] = (False, 0)

    simplify: int = 20
    merge_radius: int | None = None  # if not None will merge objects in given radius


def _industrial_danglvl_to_init(dang_lvl: int):
    init_values = {1: -10, 2: -8, 3: -6, 4: -4}
    return init_values.get(dang_lvl, -2)


def _industrial_danglvl_to_fading(dang_lvl: int):
    fading_values = {1: 0.8, 2: 0.6, 3: 0.4, 4: 0.2}
    return fading_values.get(dang_lvl, 0.1)


default_layers_options: Dict[str, LayerOptions] = {
    "industrial": LayerOptions(
        initial_impact=("dangerous_level", _industrial_danglvl_to_init),
        fading=("dangerous_level", _industrial_danglvl_to_fading),
        russian_name="Промышленный объект",
        merge_radius=250,
    ),
    "gas_station": LayerOptions(
        initial_impact=-4,
        fading=0.15,
        russian_name="АЗС",
        geom_func=lambda x: x.buffer(50),
        merge_radius=50,
    ),
    "landfill": LayerOptions(
        initial_impact=-3,
        fading=0.4,
        russian_name="Свалка",
        area_normalization=lambda x: min_max_normalization(np.sqrt(x), 1, 5, math.sqrt(1000), math.sqrt(500000)),
        merge_radius=100,
    ),
    "regional_roads": LayerOptions(
        initial_impact=-2,
        fading=0.1,
        russian_name="Дорога регионального назначения",
        geom_func=lambda x: x.buffer(20),
        geom_union_unionbuff=(True, 0),
    ),
    "federal_roads": LayerOptions(
        initial_impact=-4,
        fading=0.2,
        russian_name="Дорога федерального назначения",
        geom_func=lambda x: x.buffer(30),
        geom_union_unionbuff=(True, 0),
    ),
    "railway": LayerOptions(
        initial_impact=-3,
        fading=0.15,
        russian_name="Железнодорожные пути",
        geom_func=lambda x: x.buffer(30),
        geom_union_unionbuff=(True, 0),
    ),
    "nature_reserve": LayerOptions(
        initial_impact=4,
        fading=0.2,
        russian_name="ООПТ",
    ),
    "water": LayerOptions(
        initial_impact=3,
        fading=0.1,
        russian_name="Водный объект",
        geom_union_unionbuff=(True, 50),
    ),
    "woods": LayerOptions(
        initial_impact=3,
        fading=0.15,
        russian_name="Зелёная зона",
        make_donuts=False,
        simplify=50,
        geom_union_unionbuff=(True, 500),
    ),
}



import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
from loguru import logger
from shapely.ops import polygonize, unary_union

from ecodonut.utils import get_closest_nodes, graph_to_gdf, polygons_to_linestring, rivers_dijkstra


def simulate_spill(
    pollution_sources: gpd.GeoDataFrame,
    river_graph: nx.DiGraph,
    min_dist_to_river=100,
    pollution_column="pollution",
    return_only_polluted=True,
) -> gpd.GeoDataFrame:
    """
    Simulate the spread of pollution in a river network, based on proximity to pollution sources and pollutant levels.

    This function identifies pollution sources that are within a specified minimum distance from a river network,
    calculates how much pollution remains along each edge in the river network, and aggregates the results into
    polygons representing affected areas with pollution values.

    Args:
        pollution_sources (gpd.GeoDataFrame): A GeoDataFrame containing the locations and pollution values of sources.
        river_graph (nx.DiGraph): A directed graph representing the river network, with nodes and edges. The graph must
            contain a 'crs' attribute indicating the coordinate reference system.
        min_dist_to_river (float, optional): The maximum distance between pollution sources and river nodes for
            pollution to be considered in the simulation. Default is 100.
        pollution_column (str, optional): The name of the column in `pollution_sources` that specifies pollution values.
            Default is "pollution".
        return_only_polluted (bool, optional): Whether to return only the polluted rivers, or all rivers geometry.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing polygons representing areas affected by pollution. Each polygon
        has an attribute `remain`, which indicates the pollution concentration within that area.

    """
    if pollution_column not in pollution_sources.columns:
        raise RuntimeError(f"{pollution_column} is not in geodataframe,specify your pollution_column attribute")

    local_crs = river_graph.graph["crs"]

    pollution_sources.to_crs(local_crs, inplace=True)
    pollution_sources[["closest_node", "dist"]] = get_closest_nodes(pollution_sources, river_graph)
    pollution_sources = pollution_sources[pollution_sources["dist"] <= min_dist_to_river]
    result = pd.DataFrame()
    for _, row in pollution_sources.iterrows():
        pollution = row[pollution_column]
        weight_remain = rivers_dijkstra(river_graph, int(row.closest_node), weight="weight", cutoff=pollution)
        weight_remain = pd.DataFrame.from_dict(weight_remain, orient="index", columns=["remain"])
        result = pd.concat([result, weight_remain], axis=1)
    result = result.sum(axis=1)
    if return_only_polluted:
        subgraph = river_graph.subgraph(result.index).copy()
    else:
        subgraph = river_graph.copy()

    if subgraph.number_of_edges() == 0:
        logger.info("No pollution has found, returning empty GeoDataFrame")
        return gpd.GeoDataFrame()

    subgraph.graph["crs"] = local_crs
    for u, v, data in subgraph.edges(data=True):
        if u in result.index and v in result.index:
            data["remain"] = (result.loc[u] + result.loc[v]) / 2
        else:
            data["remain"] = 0

    edges_gdf = graph_to_gdf(subgraph)
    edges_gdf = (
        edges_gdf.groupby("remain")
        .agg({"geometry": unary_union})
        .reset_index()
        .set_geometry("geometry")
        .set_crs(local_crs)
    )
    polygons = polygonize(edges_gdf.geometry.apply(polygons_to_linestring).unary_union)
    enclosures = gpd.GeoSeries(list(polygons), crs=local_crs)
    enclosures_points = gpd.GeoDataFrame(geometry=enclosures.representative_point(), crs=enclosures.crs)

    joined = gpd.sjoin(enclosures_points, edges_gdf, how="inner", predicate="within").reset_index()

    means = joined.groupby("index")["remain"].apply(
        lambda x: float(np.mean([val for val in x if val > 0])) if np.any(np.array(x) > 0) else 0
    )
    joined = joined.drop_duplicates(subset="index")
    joined.set_index("index", inplace=True)
    joined["remain"] = means
    joined["geometry"] = enclosures
    joined = gpd.GeoDataFrame(joined, geometry="geometry", crs=local_crs).reset_index(drop=True)
    return joined



import geopandas as gpd
import pandas as pd

from ecodonut.utils import project_points_into_polygons


def industrial_preprocessing(
    gdf_dangerous_objects_points: gpd.GeoDataFrame,
    gdf_industrial_polygons: gpd.GeoDataFrame,
    dangerous_level_column="dangerous_level",
) -> gpd.GeoDataFrame:
    def remove_nulls(x):
        if isinstance(x, list):
            x = [item for item in x if pd.notnull(item)]
            if len(x) == 0:
                return None
            if len(x) == 1:
                return x[0]
        return x

    gdf_dangerous_objects_points = gdf_dangerous_objects_points.copy()
    gdf_industrial_polygons = gdf_industrial_polygons.copy()

    gdf_dangerous_objects_points["geometry"] = gdf_dangerous_objects_points.geometry
    gdf_industrial_polygons["geometry"] = gdf_industrial_polygons.geometry

    estimated_crs = gdf_industrial_polygons.estimate_utm_crs()

    gdf_dangerous_objects_points.to_crs(estimated_crs, inplace=True)

    gdf_dangerous_objects_points = gdf_dangerous_objects_points[
        gdf_dangerous_objects_points["name"]
        .str.upper()
        .str.contains("ПОЛИГОН|ЗАВОД|ТБО|ТКО|ЦЕХ|КОТЕЛЬНАЯ|РУДНИК|КАНАЛ|КАРЬЕР|СТАНЦИЯ|ПРОИЗВОД|ПРОМЫШЛЕН")
    ]

    gdf_dangerous_objects_points = gdf_dangerous_objects_points.filter(
        items=["name", "geometry", dangerous_level_column]
    )

    gdf_industrial_polygons.to_crs(estimated_crs, inplace=True)
    gdf_industrial_polygons = gdf_industrial_polygons.filter(items=["name", "geometry"])

    union = project_points_into_polygons(gdf_dangerous_objects_points, gdf_industrial_polygons)

    union[dangerous_level_column] = union.apply(lambda row: max(row["dangerous_level"]), axis=1)
    union["name"] = union["name_left"] + union["name_right"]
    union["name"] = union["name"].apply(remove_nulls)
    union["name"] = union["name"].apply(lambda x: ", ".join(set(x)) if isinstance(x, list) else x)
    union.dropna(subset="name", inplace=True)
    union.drop(columns=["name_left", "name_right", "index_right"], inplace=True)
    union["type"] = "industrial"
    union["dangerous_level"] = union["dangerous_level"].fillna(4)
    union = gpd.GeoDataFrame(union, crs=estimated_crs)
    union = union.loc[~union["geometry"].is_empty]
    union = union[~union["geometry"].duplicated()]
    return union



import geopandas as gpd
import numpy as np
import pandas as pd
from shapely import LineString, MultiLineString, MultiPolygon, Polygon
from shapely.ops import polygonize


def min_max_normalization(data, new_min=0, new_max=1, old_min=None, old_max=None):
    if old_min is None:
        old_min = np.min(data)
    if old_max is None:
        old_max = np.max(data)
    normalized_data = (data - old_min) / (old_max - old_min) * (new_max - new_min) + new_min
    return normalized_data


def calc_layer_count(gdf, minv=2, maxv=10, overall_min=None, overall_max=None) -> np.ndarray:
    impacts = np.abs(gdf["total_impact_radius"])
    if len(np.unique(impacts)) == 1:
        return np.full((impacts.shape[0]), int((minv + maxv) / 2))
    norm_impacts = min_max_normalization(impacts, minv, maxv, overall_min, overall_max)
    return np.round(norm_impacts).astype(int)


def combine_geometry(distributed, impact_calculator, agg_func=tuple) -> gpd.GeoDataFrame:
    polygons = polygonize(distributed.geometry.apply(polygons_to_linestring).unary_union)
    enclosures = gpd.GeoSeries(list(polygons), crs=distributed.crs)
    enclosures_points = gpd.GeoDataFrame(enclosures.representative_point(), columns=["geometry"], crs=enclosures.crs)

    joined = gpd.sjoin(enclosures_points, distributed, how="inner", predicate="within").reset_index()
    joined = joined.groupby("index").agg(
        {"name": agg_func, "type": agg_func, "layer_impact": tuple, "source": agg_func}
    )
    joined["layer_impact"] = joined["layer_impact"].apply(impact_calculator)
    joined["layer_impact"] = joined["layer_impact"].astype(float).apply(round, ndigits=2)
    joined["geometry"] = enclosures
    joined = gpd.GeoDataFrame(joined, geometry="geometry", crs=distributed.crs)

    return joined


def polygons_to_linestring(geom: Polygon | MultiPolygon):
    def convert_polygon(polygon: Polygon):
        lines = []
        exterior = LineString(polygon.exterior.coords)
        lines.append(exterior)
        interior = [LineString(p.coords) for p in polygon.interiors]
        lines = lines + interior
        return lines

    def convert_multipolygon(polygon: MultiPolygon):
        return MultiLineString(sum([convert_polygon(p) for p in polygon.geoms], []))

    if geom.geom_type == "Polygon":
        return MultiLineString(convert_polygon(geom))
    return convert_multipolygon(geom)


def create_buffers(loc: pd.Series, resolution, positive_func, negative_func) -> gpd.GeoDataFrame:
    layers_count = loc.layers_count
    # Calculation of each impact buffer
    radius_per_lvl = abs(round(loc.total_impact_radius / (layers_count - 1), 2))
    initial_impact = loc.initial_impact

    # Calculating the impact at each level
    each_lvl_impact = {}

    if initial_impact > 0:
        for i in range(1, layers_count):
            each_lvl_impact[i] = positive_func(layers_count, i) * initial_impact
    else:
        for i in range(1, layers_count):
            each_lvl_impact[i] = negative_func(layers_count, i) * initial_impact

    loc["layer_impact"] = initial_impact
    loc["source"] = True

    # Creating source level
    distributed_df = pd.DataFrame()
    distributed_df = pd.concat(
        [distributed_df, loc[["name", "type", "layer_impact", "source", "geometry"]].to_frame().T]
    )

    initial_geom = loc["geometry"]
    to_cut = initial_geom
    new_layer = loc.copy()

    radius = 0

    for i in range(1, layers_count):
        radius = radius + radius_per_lvl
        impact = each_lvl_impact.get(i)
        new_layer["layer_impact"] = round(impact, 2)
        new_layer["source"] = False
        loc_df = gpd.GeoDataFrame(
            new_layer[["name", "type", "layer_impact", "source", "geometry"]].to_frame().T, geometry="geometry"
        )
        to_cut_temp = initial_geom.buffer(radius, resolution=resolution)
        loc_df["geometry"] = to_cut_temp.difference(to_cut)
        to_cut = to_cut_temp
        distributed_df = pd.concat([distributed_df, loc_df])

    return gpd.GeoDataFrame(distributed_df, geometry="geometry").reset_index(drop=True)


def merge_objs_by_buffer(gdf: gpd.GeoDataFrame, buffer: int) -> gpd.GeoDataFrame:
    """
    Function to merge geometries based on a specified buffer.

    Parameters
    ----------
    gdf: gpd.GeoDataFrame
        A GeoPandas GeoDataFrame that contains the geometries to be merged.
    buffer: int
        Buffer value for merging objects.

    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame containing the merged geometries. Attributes are preserved in lists with unique values.

    Examples
    --------
    >>> gdf = gpd.read_file('path_to_your_file.geojson')
    >>> result = merge_objs_by_buffer(gdf, buffer=100)
    """

    def unique_list(data):
        if len(tuple(data)) == 1:
            return data
        return tuple(data)

    crs = gdf.crs
    buffered = gpd.GeoDataFrame(geometry=[Polygon(x) for x in gdf.buffer(buffer).union_all().geoms], crs=crs)
    representative_points = gdf.copy()
    representative_points.geometry = representative_points.geometry.representative_point()
    joined = gpd.sjoin(representative_points, buffered, how="inner", predicate="within")
    joined.geometry = gdf.geometry
    joined = joined.groupby("index_right").agg({x: unique_list for x in gdf.columns})
    joined.geometry = joined.geometry.apply(lambda x: gpd.GeoSeries(x).union_all())
    joined = gpd.GeoDataFrame(joined, geometry="geometry", crs=crs).reset_index(drop=True)
    return joined


def project_points_into_polygons(
    points: gpd.GeoDataFrame, polygons: gpd.GeoDataFrame, polygons_buff=10
) -> gpd.GeoDataFrame:
    assert points.crs == polygons.crs, "Non-matched crs"
    assert points.crs != 4326, "Geographical crs"
    assert polygons.crs != 4326, "Geographical crs"

    polygons_buffered = polygons.copy()
    polygons_buffered["geometry"] = polygons_buffered.geometry.buffer(polygons_buff)

    intersect = gpd.sjoin(polygons_buffered, points, how="left").reset_index()
    intersect = intersect.groupby("index").agg(list)
    intersect["geometry"] = polygons.loc[intersect.index.values, "geometry"]
    return intersect
